from __future__ import annotations

from .. import consts
from .anime import Anime

from functools import cached_property
from typing import TYPE_CHECKING, Iterator, Literal, Callable

if TYPE_CHECKING:
    from ..core import Core


class Database:
    '''
    Represents the website database.
    '''
    
    def __init__(self, core: Core) -> None:
        '''
        Initialise a new database instance.
        
        :param core: Nekosama core to use. 
        '''
        
        self.core = core
    
    def _fetch(self, url: str) -> list[dict]:
        '''
        Shorthand for fetching a JSON ressource.
        
        :param url: Ressource URL to fetch.
        :return: The parsed ressource.
        '''
        
        response = self.core.session.get(url)
        response.raise_for_status()
        return response.json()
    
    @cached_property
    def VO(self) -> list[dict]:
        return self._fetch('https://neko-sama.fr/animes-search-vostfr.json')
    
    @cached_property
    def VF(self) -> list[dict]:
        return self._fetch('https://neko-sama.fr/animes-search-vf.json')
    
    def search(self,
               query: str = None,
               lang: Literal['VO', 'VF'] = 'VO',
               type: consts.Type = None,
               genres: set[consts.Genre] = None,
               custom: Callable[[dict], bool] = None) -> Iterator[Anime]:
        '''
        Search for animes in the database.
        
        :param query: Keywords to search for.
        :param lang: Database language.
        :param type: Specific anime type.
        :param genres: Specific anime genres.
        :param custom: A custom filter callable.
        :return: An Anime iterator.
        '''
        
        def callback(data: dict) -> bool | None:
            
            if custom and not custom(data):                           return False
            if type   and type != data['type']:                       return False
            if genres and not genres.issubset(data['genres']):        return False
            if query  and query.lower() not in data['title'].lower(): return False
            
            return True
        
        for data in filter(callback, getattr(self, lang)):
            yield Anime(self.core, data['url'])

# EOF


from __future__ import annotations

import os
from concurrent import futures
from rich.progress import Progress
from functools import cached_property
from typing import TYPE_CHECKING, Literal, TypedDict

from .. import consts
from .episode import Episode

if TYPE_CHECKING:
    from ..core import Core


class AnimeData(TypedDict):
    title: str
    type: str
    url: str
    image: str
    description: str


class Anime:
    '''
    Represents an anime.
    '''
    
    def __init__(self, core: Core, url: str) -> None:
        '''
        Initialises a new Anime object.
        
        :param core: Nekosama core to use.
        :param url: Anime URL.
        '''
        
        self.core = core
        self.url = url.lower()
        
        url_data = consts.re.url_data.match(self.url).groups()
        self.id = int(url_data[0])
        self.slug: str = url_data[1]
        self.lang: Literal['VO', 'VF'] = url_data[2][:2].upper()
        
    def __repr__(self) -> str:
        return f'Anime({self.slug} {self.lang})'
    
    @cached_property
    def page(self) -> str:
        '''
        The anime HTML source.
        '''

        response = self.core.session.get(self.url)
        response.raise_for_status()
        return response.text
    
    @cached_property
    def data(self) -> AnimeData:
        '''
        The anime data.
        '''
        
        return dict(consts.re.anime_data.findall(self.page))
    
    @cached_property
    def title(self) -> str:
        '''
        The anime title.
        '''
        
        return self.data['title'].split('|')[0].strip()
    
    def get_image(self) -> bytes:
        '''
        Get the anime poster.
        
        :return: The raw image bytes.
        '''
        
        response = self.core.session.get(self.data['image'])
        response.raise_for_status()
        return response.content
    
    @cached_property
    def episodes(self) -> list[Episode]:
        '''
        Anime episodes.
        '''
        
        urls = consts.re.episodes.findall(self.page)
        return [Episode(self, url, i) for i, url in enumerate(urls, start = 1)]

    def download(self,
                 path: str,
                 quality: Literal[1080, 720, 480] = 1080,
                 workers: int = 1) -> None:
        '''
        Downloads the whole anime.
        
        :param path: Parent directory for all downloads.
        :param quality: Download quality.
        :param workers: Amount of concurent download workers.
        '''
        
        os.makedirs(path, exist_ok = True)
        
        with Progress() as progress, futures.ThreadPoolExecutor(workers) as pool:
            
            threads = [
                pool.submit(
                    episode.download,
                    path = os.path.join(path, f'{episode.index}.mp4'),
                    quality = quality,
                    tracker = progress
                )
                for episode in self.episodes
            ]
            
            
            for thread in threads:
                if thread.cancelled():
                    continue
                
                try:
                    thread.result()
                
                except KeyboardInterrupt:
                    pool.shutdown(wait = False, cancel_futures = True)
                    raise

# EOF


from __future__ import annotations

import base64
from yt_dlp import YoutubeDL
from rich.progress import Progress
from contextlib import nullcontext
from functools import cached_property
from typing import TYPE_CHECKING, Literal, Callable, Type

from .. import consts

if TYPE_CHECKING:
    from .anime import Anime


class Episode:
    '''
    Represents an anime episode.
    '''
    
    def __init__(self, anime: Anime, url: str, index: int) -> None:
        '''
        Initialises a new Episode object.
        
        :param anime: Parent anime the episode is issued from.
        :param url: The episode URL.
        :param index: The episode index (1-based).
        '''
        
        self.anime = anime
        self.core = anime.core
        self.url = url
        self.index: int = index
    
    def __repr__(self) -> str:
        return f'Episode({self.anime.slug} {self.anime.lang} - E{self.index})'
    
    @cached_property
    def page(self) -> str:
        '''
        The episode HTML source.
        '''
        
        response = self.core.session.get(self.url)
        response.raise_for_status()
        return response.text
    
    def get_hls(self,
                quality: Literal[1080, 720, 480] = 1080,
                player: int = 0) -> str:
        '''
        Get the HLS playlist URL.
        
        :param quality: The video quality of the source.
        :param player: Player source index to use.
        :return: A valid HLS URL. Will expire after some time.
        '''
        
        html = self.core.session.get(consts.re.players.findall(self.page)[player]).text
        js   = self.core.session.get(consts.re.script.findall(html)[0]).text
        data = base64.b64decode(consts.re.atob.findall(js)[0]).decode()
        hls  = self.core.session.get(consts.re.url.findall(data)[0].replace('\\', '')).text
        return dict(consts.re.qualities.findall(hls))[str(quality)]
    
    def _download(self,
                  path: str,
                  url: str,
                  callback: Callable[[int, int], None],
                  **dl_kw) -> None:
        '''
        Internal download backend.
        
        :param path: The out-tmpl YTDLP arg.
        :param url: the HLS source URL.
        :param callback: Callable for download tracking.
        :param **dl_kw: Additional YTDLP arguments. 
        '''
        
        def hook(data: dict) -> None:
            callback(
                data.get('downloaded_bytes', 0),
                data.get('total_bytes_estimate', 0)
            )
        
        with YoutubeDL({
            'outtmpl': path,
            'progress_hooks': [hook],
            'http_headers': consts.headers,
            'retries': 5,
            'quiet': True,
            'nopart': True,
            'noprogress': True,
            'no_warnings': True
        } | dl_kw) as ytdl:
            ytdl.download([url])
    
    def download(self,
                 path: str,
                 quality: Literal[1080, 720, 480] = 1080,
                 callback: Callable[[int, int], None] = None,
                 tracker: Progress | Type[Progress] = Progress,
                 **dl_kw) -> None:
        '''
        Downloads the episode.
        
        :param path: The output file path.
        :param quality: The video source quality.
        :param callback: Callable to track download progress.
        :param tracker: rich progress display for simultaneous downloads.
        :param **dl_kw: Additional YTDLP arguments.
        '''
        
        with (tracker() if tracker == Progress else nullcontext(tracker)) as progress:
            
            task = progress.add_task(f'E{self.index}')
            
            def wrapper(cur: int, total: int) -> None:
                progress.update(task, completed = cur, total = total)
                if callback: callback(cur, total)
            
            self._download(
                path = path,
                url = self.get_hls(quality),
                callback = wrapper,
                dl_kw = dl_kw
            )

# EOF


from re import compile
from typing import Literal

headers = {
    'Accept': '*/*',
    'Accept-Language': 'en,en-US',
    'Connection': 'keep-alive',
    'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64; rv:2.0.1) Gecko/20100101 Firefox/4.0.1'
}

class re:
    '''
    Collection of internal regexes.
    '''
    
    anime_data = compile(r'<meta property=\"og:(?P<name>.*)\" content=\"([.\s\S]*?)\" />')
    url_data   = compile(r'https://.*?/(\d*)-([a-z-\d]*)_(vostfr|vf)')
    episodes   = compile(r'https://neko-sama\.fr/anime/episode/\d*[a-z-\d_]*')
    players    = compile(r"video\[\d\] = '(.*?)';")
    script     = compile(r'https://[a-z.]*\/[a-z\/-]*\?.*\" ')
    atob       = compile(r'atob\(\"(.*?)\"\)')
    url        = compile(r'(https.*?)\"')
    qualities  = compile(r'=\"(\d+?)\"\n(http.*?)(?:\n|$)')

Type = Literal['tv', 'ova', 'm0v1e', 'special']

Genre = Literal[
    'Psychological', 'Action', 'Hentai', 'Drama', 'Mahou Shoujo', 'Music', 'Mecha', 'Yuri',
    'Adventure', 'Shoujo', 'Shounen', 'Romance', 'Ecchi', 'Sports', 'Mafia', 'Battle Royale',
    'Mystery', 'Thriller', 'Magic', 'Horror', 'Comedy', 'Military', 'Isekai', 'Slice of Life',
    'Supernatural', 'Sci-Fi', 'Fantasy', 'Cyberpunk'
]

# EOF


import os
import argparse

from nekosama import Core


header = '\x1b[2m( NEKO )\x1b[0m'

parser = argparse.ArgumentParser(
    prog = 'nekosama',
    description = 'Download CLI for nekosama.fr\nSee https://github.com/Egsagon/neko-sama-api for help.'
)

def error(message: str) -> None:
    print(header, f'\x1b[91m{message}\x1b[0m')
    exit()

parser.error = error

parser.add_argument(
    '-i', '--input',
    help = 'An anime URL',
    required = True
)

parser.add_argument(
    '-e', '--episodes',
    help = 'List or range or episodes indexes (e.g. 0-5 or 1,2,3)',
    default = '0-'
)

parser.add_argument(
    '-o', '--output',
    help = 'Output directory',
    default = '.'
)

parser.add_argument(
    '-q', '--quality',
    help = 'Video quality',
    default = '1080p',
    choices = ['1080p', '720p', '480p']
)

args = parser.parse_args()

# Get anime
core = Core()
anime = core.get(args.input)

print(header, f'Fetched anime \x1b[92m{anime.title}\x1b[0m')

# Get episode range
if ',' in args.episodes:
    indexes = list(map(int, args.episodes.split(',')))

elif '-' in args.episodes:    
    start, stop = args.episodes.split('-')    
    indexes = range(int(start or 0), int(stop or -1) + 1)

else:
    indexes = [int(args.episodes)]

episodes = [ep for ep in anime.episodes if ep.index in indexes]

if not episodes:
    error('No episode found matching query')

print(header, f'Downloading \x1b[92m{len(episodes)}\x1b[0m episodes.')

try:
    for episode in episodes:
        episode.download(
            path = os.path.join(args.output, f'{episode.anime.slug}-ep{episode.index}'),
            quality = int(args.quality.replace('p', ''))
        )

except KeyboardInterrupt:
    error('Canceled by user')

except Exception as err:
    error(repr(err))

print(header, 'Success!')

# EOF


import httpx

from . import consts
from .objects.anime import Anime
from .objects.database import Database


class Core:
    '''
    Represents a single connection to the target website.
    '''
    
    def __init__(self, client: httpx.Client = None) -> None:
        '''
        Initialises a new Core object.
        
        :param client: Custom HTTPX client to use for requests.
        '''
        
        self.session = client or httpx.Client(
            headers = consts.headers,
            follow_redirects = True
        )
        
        self.database = Database(self)
    
    def get(self, url: str) -> Anime:
        '''
        Get an anime or episode.
        
        :param url: The Anime URL.
        :return: A corresponding Anime object. 
        '''
        
        return Anime(self, url)

# EOF


'''
    Nekosama 4.0
    
    See https://github.com/Egsagon/neko-sama-api.
'''

from .core import Core
from . import consts

from .objects.anime import Anime
from .objects.episode import Episode
from .consts import Type, Genre

from rich.progress import Progress

__all__ = ['Core', 'consts', 'Anime', 'Episode', 'Type', 'Genre']

# EOF